// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AISS16AO2/driver/device.c $
// $Rev: 52935 $
// $Date: 2023-04-20 15:51:33 -0500 (Thu, 20 Apr 2023) $

// 16AISS16AO2: Device Driver: source file

#include "main.h"



// variables ******************************************************************

const gsc_dev_id_t	dev_id_list[]	=
{
	// model		Vendor	Device	SubVen	SubDev	type
	{ "16AISS16AO2",	0x10B5, 0x9056, 0x10B5, 0x3243,	GSC_DEV_TYPE_16AISS16AO2	},
	{ NULL }
};



//*****************************************************************************
static void _channels_compute(dev_data_t* dev)
{
	if (dev->cache.gsc_acr_32 & 0x10000)
		dev->cache.ai_chan_qty	= 8;
	else
		dev->cache.ai_chan_qty	= 16;

	if (dev->cache.gsc_acr_32 & 0x20000)
		dev->cache.ao_chan_qty	= 0;
	else
		dev->cache.ao_chan_qty	= 2;

	dev->cache.ai_chans_max	= 16;
	dev->cache.ao_chans_max	= 2;
}



//*****************************************************************************
static void _master_clock_compute(dev_data_t* dev)
{
	switch (dev->cache.gsc_acr_32 & 0xC0000)
	{
		case 0x00000:	dev->cache.master_clock	= 45000000L;	break;
		case 0x40000:
		case 0x80000:
		case 0xC0000:
		default:		dev->cache.master_clock	= -1;			break;
	}

	dev->cache.rate_gen_ndiv_min	= dev->cache.master_clock / dev->cache.fsamp_max;
}



//*****************************************************************************
static void _dmdma_compute(dev_data_t* dev)
{
	u16	fw	= dev->cache.gsc_acr_32 & 0xFFF;

	dev->cache.dmdma	= (fw >= 0x006) ? 1 : 0;
}



/******************************************************************************
*
*	Function:	dev_device_create
*
*	Purpose:
*
*		Do everything needed to setup and use the given device.
*
*	Arguments:
*
*		dev		The structure to initialize.
*
*	Returned:
*
*		0		All is well.
*		< 0		An appropriate error status.
*
******************************************************************************/

int dev_device_create(dev_data_t* dev)
{
	static const gsc_bar_maps_t	bar_map	=
	{
		{
			// mem	io	rw
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 0: PLX registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 1: PLX registers, I/O mapped
			{ 1,	0,	GSC_REG_TYPE_ACCESS_RW },	// BAR 2: GSC registers, memory mapped
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 3: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO },	// BAR 4: unused
			{ 0,	0,	GSC_REG_TYPE_ACCESS_RO }	// BAR 5: unused
		}
	};

	u32	dma0;
	u32	dma1;
	int	ret;

	for (;;)	// A convenience loop.
	{
		// Verify some macro contents.
		ret	= gsc_macro_test_base_name(AISS16AO2_BASE_NAME);
		if (ret)	break;

		ret	= gsc_macro_test_model();
		if (ret)	break;

		// PCI setup.
		ret	= os_pci_dev_enable(&dev->pci);
		if (ret)	break;

		ret	= os_pci_master_set(&dev->pci);
		if (ret)	break;

		// Control ISR access to the device and data structure.
		ret	= os_spinlock_create(&dev->spinlock);
		if (ret)	break;

		// Control access to the device and data structure.
		ret	= os_sem_create(&dev->sem);
		if (ret)	break;

		// Access the BAR regions.
		ret	= gsc_bar_create(dev, &dev->bar, &bar_map);
		if (ret)	break;

		// Firmware access.
		dev->vaddr.gsc_acr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_ACR));
		dev->vaddr.gsc_aibbsr_32	= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_AIBBSR));
		dev->vaddr.gsc_aibr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_AIBR));
		dev->vaddr.gsc_aicr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_AICR));
		dev->vaddr.gsc_aobr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_AOBR));
		dev->vaddr.gsc_bcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_BCR));
		dev->vaddr.gsc_boor_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_BOOR));
		dev->vaddr.gsc_diopr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_DIOPR));
		dev->vaddr.gsc_ibsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_IBSR));
		dev->vaddr.gsc_ibtr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_IBTR));
		dev->vaddr.gsc_obsr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_OBSR));
		dev->vaddr.gsc_obtr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_OBTR));
		dev->vaddr.gsc_psr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_PSR));
		dev->vaddr.gsc_rgar_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_RGAR));
		dev->vaddr.gsc_rgbr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_RGBR));
		dev->vaddr.gsc_rgcr_32		= GSC_VADDR(dev, GSC_REG_OFFSET(AISS16AO2_GSC_RGCR));

		// Data cache initialization.
		dev->cache.gsc_acr_32			= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_acr_32);

		dev->cache.auto_cal_ms			= 10000;
		dev->cache.fifo_size_rx			= _256K;
		dev->cache.fifo_size_tx			= _256K;
		dev->cache.fsamp_max			= 1000000L;
		dev->cache.fsamp_min			= 3;	// 45MHz / 0xFFFFFF
		dev->cache.initialize_ms		= 3;
		dev->cache.rate_gen_fgen_max	= 1000000L;
		dev->cache.rate_gen_fgen_min	= 3;	// 45MHz / 0xFFFFFF
		dev->cache.rate_gen_ndiv_mask	= 0xFFFFFF;
		dev->cache.rate_gen_ndiv_max	= 0xFFFFFF;
		dev->cache.rate_gen_ndiv_min	= 45;
		dev->cache.rate_gen_qty			= 3;

		_channels_compute(dev);
		_master_clock_compute(dev);	// after fsamp
		_dmdma_compute(dev);

		// Initialize additional resources.
		ret	= dev_irq_create(dev);
		if (ret)	break;

		ret	= dev_io_create(dev);
		if (ret)	break;

		dma0	= GSC_DMA_SEL_STATIC
				| GSC_DMA_CAP_BMDMA_READ;
		dma1	= GSC_DMA_SEL_STATIC
				| GSC_DMA_CAP_BMDMA_WRITE;

		if (dev->cache.dmdma)
		{
			dma0	|= GSC_DMA_CAP_DMDMA_READ;
			dma1	|= GSC_DMA_CAP_DMDMA_WRITE;
		}

		ret	= gsc_dma_create(dev, dma0, dma1);
		break;
	}

	return(ret);
}



/******************************************************************************
*
*	Function:	dev_device_destroy
*
*	Purpose:
*
*		Do everything needed to release the referenced device.
*
*	Arguments:
*
*		dev		The partial data for the device of interest.
*
*	Returned:
*
*		None.
*
******************************************************************************/

void dev_device_destroy(dev_data_t* dev)
{
	if (dev)
	{
		gsc_dma_destroy(dev);
		dev_io_destroy(dev);
		dev_irq_destroy(dev);
		gsc_bar_destroy(&dev->bar);
		os_sem_destroy(&dev->sem);
		os_spinlock_destroy(&dev->spinlock);
		os_pci_master_clear(&dev->pci);
		os_pci_dev_disable(&dev->pci);
	}
}


