// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/16AISS16AO2/driver/ioctl.c $
// $Rev: 52935 $
// $Date: 2023-04-20 15:51:33 -0500 (Thu, 20 Apr 2023) $

// 16AISS16AO2: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTO_CAL_START			D28			// BCR
#define	AUTO_CAL_PASS			D29			// BCR
#define	INIT_START				D31			// BCR

#define	IRQ_AUTO_CAL_DONE		0x00000001L	// PSR



//*****************************************************************************
static int _query(dev_data_t* dev, s32* arg)
{
	switch (arg[0])
	{
		default:								arg[0]	= AISS16AO2_IOCTL_QUERY_ERROR;		break;
		case AISS16AO2_QUERY_AUTO_CAL_MS:		arg[0]	= dev->cache.auto_cal_ms;			break;
		case AISS16AO2_QUERY_CHANNEL_AI_MAX:	arg[0]	= dev->cache.ai_chans_max;			break;
		case AISS16AO2_QUERY_CHANNEL_AI_QTY:	arg[0]	= dev->cache.ai_chan_qty;			break;
		case AISS16AO2_QUERY_CHANNEL_AO_MAX:	arg[0]	= dev->cache.ao_chans_max;			break;
		case AISS16AO2_QUERY_CHANNEL_AO_QTY:	arg[0]	= dev->cache.ao_chan_qty;			break;
		case AISS16AO2_QUERY_COUNT:				arg[0]	= AISS16AO2_IOCTL_QUERY_LAST + 1;	break;
		case AISS16AO2_QUERY_DEVICE_TYPE:		arg[0]	= dev->board_type;					break;
		case AISS16AO2_QUERY_DMDMA:				arg[0]	= dev->cache.dmdma;					break;
		case AISS16AO2_QUERY_FGEN_MAX:			arg[0]	= dev->cache.rate_gen_fgen_max;		break;
		case AISS16AO2_QUERY_FGEN_MIN:			arg[0]	= dev->cache.rate_gen_fgen_min;		break;
		case AISS16AO2_QUERY_FIFO_SIZE_RX:		arg[0]	= dev->cache.fifo_size_rx;			break;
		case AISS16AO2_QUERY_FIFO_SIZE_TX:		arg[0]	= dev->cache.fifo_size_tx;			break;
		case AISS16AO2_QUERY_FSAMP_MAX:			arg[0]	= dev->cache.fsamp_max;				break;
		case AISS16AO2_QUERY_FSAMP_MIN:			arg[0]	= dev->cache.fsamp_min;				break;
		case AISS16AO2_QUERY_INIT_MS:			arg[0]	= dev->cache.initialize_ms;			break;
		case AISS16AO2_QUERY_MASTER_CLOCK:		arg[0]	= dev->cache.master_clock;			break;
		case AISS16AO2_QUERY_NDIV_MASK:			arg[0]	= dev->cache.rate_gen_ndiv_mask;	break;
		case AISS16AO2_QUERY_NDIV_MAX:			arg[0]	= dev->cache.rate_gen_ndiv_max;		break;
		case AISS16AO2_QUERY_NDIV_MIN:			arg[0]	= dev->cache.rate_gen_ndiv_min;		break;
		case AISS16AO2_QUERY_RATE_GEN_QTY:		arg[0]	= dev->cache.rate_gen_qty;			break;
	}

	return(0);
}



//*****************************************************************************
int initialize_ioctl(dev_data_t* dev, void* arg)
{
	u32		bcr;
	int		i;
	long	ms		= dev->cache.initialize_ms + 5000;
	int		ret		= 0;
	int		tmp;

	// This device doesn't support an INITIALIZE DONE interrupt.

	// Initiate initialization.
	os_reg_mem_tx_u32(dev, dev->vaddr.gsc_bcr_32, INIT_START);

	// Manually wait for completion.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bcr_32, INIT_START, 0);

	if (tmp)
	{
		ret	= ret ? ret : tmp;
		printf(	"%s: INITIALIZATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (bcr & INIT_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: INITILIZE STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Initialize the software settings.

	for (i = 0; i < DEV_IO_STREAM_QTY; i++)
	{
		if (dev->io.io_streams[i])
		{
			if (dev->io.io_streams[i]->dev_io_sw_init)
			{
				(dev->io.io_streams[i]->dev_io_sw_init)(dev, dev->io.io_streams[i]);
			}
		}
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_start(dev_data_t* dev, void* arg)
{
	// Initiate auto-calibration.
	os_reg_mem_mx_u32(NULL, dev->vaddr.gsc_bcr_32, AUTO_CAL_START, AUTO_CAL_START);
	return(0);
}



//*****************************************************************************
static int _auto_calibrate(dev_data_t* dev, void* arg)
{
	u32			mask;
	long		ms		= dev->cache.auto_cal_ms + 5000;
	u32			reg;
	int			ret;
	os_sem_t	sem;
	int			tmp;
	u32			value;
	gsc_wait_t	wt;

	// Safely select the Auto-Cal Done interrupt.
	mask	= IRQ_AUTO_CAL_DONE;
	value	= IRQ_AUTO_CAL_DONE;
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_psr_32, value, mask);

	// Wait for the local interrupt.
	os_sem_create(&sem);	// dummy, required for wait operations.
	memset(&wt, 0, sizeof(wt));
	wt.flags		= GSC_WAIT_FLAG_INTERNAL;
	wt.gsc			= AISS16AO2_WAIT_GSC_AUTO_CAL_DONE;
	wt.timeout_ms	= ms;
	ret				= gsc_wait_event(dev, &wt, _auto_cal_start, NULL, &sem);
	os_sem_destroy(&sem);

	if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATE IRQ TIMED OUT AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually wait for completion in case something terminates our wait early.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bcr_32, AUTO_CAL_START, 0);

	if (tmp)
	{
		ms	*= 2;
		ret	= ret ? ret : tmp;
		printf(	"%s: AUTOCALIBRATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (reg & AUTO_CAL_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATION STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Final results.
	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if ((reg & AUTO_CAL_PASS) == 0)
	{
		ret	= ret ? ret : -EIO;
		printf(	"%s: AUTOCALIBRATION FAILED (%ld ms).\n",
				dev->model,
				ms);
	}

	if (ret == 0)
	{
		// Wait for settling.
		os_time_sleep_ms(100);
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_sts(dev_data_t* dev, s32* arg)
{
	u32	reg;

	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

	if (reg & AUTO_CAL_START)
		arg[0]	= AISS16AO2_AUTO_CAL_STS_ACTIVE;
	else if ((reg & AUTO_CAL_PASS) == 0)
		arg[0]	= AISS16AO2_AUTO_CAL_STS_FAIL;
	else
		arg[0]	= AISS16AO2_AUTO_CAL_STS_PASS;

	return(0);
}



//*****************************************************************************
static int _rx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		GSC_IO_MODE_DMDMA,
		-1
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, list, &dev->io.rx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _rx_io_overflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_IO_OVERFLOW_IGNORE,
		AISS16AO2_IO_OVERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.overflow_check);
	return(ret);
}



//*****************************************************************************
static int _rx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AISS16AO2_IO_TIMEOUT_MIN,
			AISS16AO2_IO_TIMEOUT_INFINITE,
			&dev->io.rx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _rx_io_underflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_IO_UNDERFLOW_IGNORE,
		AISS16AO2_IO_UNDERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.underflow_check);
	return(ret);
}



//*****************************************************************************
static int _tx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		GSC_IO_MODE_DMDMA,
		-1
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, list, &dev->io.tx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _tx_io_overflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_IO_OVERFLOW_IGNORE,
		AISS16AO2_IO_OVERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.tx.overflow_check);
	return(ret);
}



//*****************************************************************************
static int _tx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			AISS16AO2_IO_TIMEOUT_MIN,
			AISS16AO2_IO_TIMEOUT_INFINITE,
			&dev->io.tx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_clear(dev_data_t* dev, void* arg)
{
	u32	bcr;
	int	i;
	int	ret		= 0;

	// Clear the buffer.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, D13, D13);

	// Wait for the bit to clear.

	for (i = 0;; i++)
	{
		bcr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bcr_32);

		if ((bcr & D13) == 0)
			break;

		if (i >= 250)
		{
			ret	= -EINVAL;
			printf(	"%s: The analog input buffer took too long to clear.\n",
					DEV_NAME);
			break;
		}
	}

	// Clear the Overflow status bit.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, 0, D15);

	// Clear the Underflow status bit.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, 0, D3);

	return(ret);
}



//*****************************************************************************
static int _ai_buf_level(dev_data_t* dev, void* arg)
{
	u32*	ptr	= arg;

	ptr[0]	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_ibsr_32);
	ptr[0]	&= 0x7FFFF;
	return(0);
}



//*****************************************************************************
static int _ai_buf_overflow(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_BUF_OVERFLOW_NO,
		AISS16AO2_BUF_OVERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 15, 15);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_underflow(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_BUF_UNDERFLOW_NO,
		AISS16AO2_BUF_UNDERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 3, 3);
	return(ret);
}



//*****************************************************************************
static int _ai_burst_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_BURST_ENABLE_NO,
		AISS16AO2_BURST_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 9, 9);
	return(ret);
}



//*****************************************************************************
static int _ai_burst_size(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFFFF, dev->vaddr.gsc_aibbsr_32, 23, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_burst_status(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_AI_BURST_STATUS_BUSY,
		AISS16AO2_AI_BURST_STATUS_READY,
		-1	// terminate list
	};

	int		ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _ai_chan_sel(dev_data_t* dev, void* arg)
{
	int	ret;

	if (dev->cache.ai_chan_qty == 16)
		ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFF, dev->vaddr.gsc_aicr_32, 15, 0);
	else
		ret	= gsc_s32_range_reg(dev, arg, 0, 0x00FF, dev->vaddr.gsc_aicr_32,  7, 0);

	return(ret);
}



//*****************************************************************************
static int _ai_clock_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_CLOCK_MODE_IN,
		AISS16AO2_CLOCK_MODE_OUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 24, 24);
	return(ret);
}



//*****************************************************************************
static int _ai_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_AIO_ENABLE_NO,
		AISS16AO2_AIO_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 12, 12);
	return(ret);
}



//*****************************************************************************
static int _ai_fsamp_range(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_FSAMP_RANGE_LOW,
		AISS16AO2_FSAMP_RANGE_HIGH,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 30, 30);
	return(ret);
}



//*****************************************************************************
static int _ai_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_AI_MODE_DIFF,
		AISS16AO2_AI_MODE_SINGLE,
		AISS16AO2_AI_MODE_ZERO,
		AISS16AO2_AI_MODE_VREF,
		AISS16AO2_AI_MODE_AO_0,
		AISS16AO2_AI_MODE_AO_1,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 2, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_range_a(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_RANGE_2_5V,
		AISS16AO2_RANGE_5V,
		AISS16AO2_RANGE_10V,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 5, 4);
	return(ret);
}



//*****************************************************************************
static int _ai_range_b(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_RANGE_2_5V,
		AISS16AO2_RANGE_5V,
		AISS16AO2_RANGE_10V,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 7, 6);
	return(ret);
}



//*****************************************************************************
static int _ai_sw_clock(dev_data_t* dev, void* arg)
{
	#undef	MASK
	#undef	SHIFT
	#define	MASK	(0x1L << SHIFT)
	#define	SHIFT	8

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, MASK, MASK);
	return(0);
}



//*****************************************************************************
static int _ai_sw_trigger(dev_data_t* dev, void* arg)
{
	#undef	MASK
	#undef	SHIFT
	#define	MASK	(0x1L << SHIFT)
	#define	SHIFT	11

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, MASK, MASK);
	return(0);
}



//*****************************************************************************
static int _ai_thresh_lvl(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x7FFFF, dev->vaddr.gsc_ibtr_32, 18, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_thresh_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_THRESH_STS_CLEAR,
		AISS16AO2_THRESH_STS_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_ibtr_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_clear(dev_data_t* dev, void* arg)
{
	u32	boor;
	int	i;
	int	ret		= 0;

	// Clear the buffer.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, D11, D11);

	// Wait for the bit to clear.

	for (i = 0;; i++)
	{
		boor	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_boor_32);

		if ((boor & D11) == 0)
			break;

		if (i >= 250)
		{
			ret	= -EINVAL;
			printf(	"%s: The analog input buffer took too long to clear.\n",
					DEV_NAME);
			break;
		}
	}

	// Clear the Data Overflow status bit.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, 0, D16);

	// Clear the Frame Overflow status bit.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, 0, D17);

	return(ret);
}



//*****************************************************************************
static int _ao_load_req(dev_data_t* dev, void* arg)
{
	#undef	MASK
	#undef	SHIFT
	#define	MASK	(0x1L << SHIFT)
	#define	SHIFT	9

	int	ret	= 0;

	// Make the request.
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_boor_32, MASK, MASK);

	// Wait for the request to be granted.

	if (dev->io.tx.timeout_s)
	{
		ret	= gsc_poll_u32(dev, dev->io.tx.timeout_s * 1000, dev->vaddr.gsc_boor_32, 0x400, 0x400);
	}

	return(ret);
}



//*****************************************************************************
static int _ao_load_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_BUF_LOAD_STS_NOT_READY,
		AISS16AO2_BUF_LOAD_STS_READY,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 10, 10);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_BUF_MODE_OPEN,
		AISS16AO2_BUF_MODE_CIRC,
		-1	// terminate list
	};

	int		ret;

	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 8, 8);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_level(dev_data_t* dev, void* arg)
{
	u32*	ptr	= arg;

	ptr[0]	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_obsr_32);
	ptr[0]	&= 0x7FFFF;
	return(0);
}



//*****************************************************************************
static int _ao_buf_over_data(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_BUF_OVERFLOW_NO,
		AISS16AO2_BUF_OVERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _ao_buf_over_fram(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_BUF_OVERFLOW_NO,
		AISS16AO2_BUF_OVERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 17, 17);
	return(ret);
}



//*****************************************************************************
static int _ao_burst_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_BURST_ENABLE_NO,
		AISS16AO2_BURST_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _ao_burst_status(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_AO_BURST_STATUS_BUSY,
		AISS16AO2_AO_BURST_STATUS_READY,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _ao_chan_sel(dev_data_t* dev, void* arg)
{
	int	ret;

	if (dev->cache.ao_chan_qty == 2)
		ret	= gsc_s32_range_reg(dev, arg, 0, 0x3, dev->vaddr.gsc_boor_32, 1, 0);
	else
		ret	= -EINVAL;

	return(ret);
}



//*****************************************************************************
static int _ao_clock_status(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_CLOCK_STATUS_NOT_READY,
		AISS16AO2_CLOCK_STATUS_READY,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 6, 6);
	return(ret);
}



//*****************************************************************************
static int _ao_clock_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_CLOCK_MODE_IN,
		AISS16AO2_CLOCK_MODE_OUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 4, 4);
	return(ret);
}



//*****************************************************************************
static int _ao_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_AIO_ENABLE_NO,
		AISS16AO2_AIO_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_boor_32, 5, 5);
	return(ret);
}



//*****************************************************************************
static int _ao_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_AO_MODE_REG,
		AISS16AO2_AO_MODE_FIFO,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _ao_output_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_AO_OUTPUT_MODE_IMMED,
		AISS16AO2_AO_OUTPUT_MODE_SIMUL,
		-1	// terminate list
	};

	int	ret;

	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _ao_range(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_RANGE_2_5V,
		AISS16AO2_RANGE_5V,
		AISS16AO2_RANGE_10V,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 17, 16);
	return(ret);
}



//*****************************************************************************
static int _ao_sw_clock(dev_data_t* dev, void* arg)
{
	#undef	MASK
	#undef	SHIFT
	#define	MASK	(0x1L << SHIFT)
	#define	SHIFT	20

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, MASK, MASK);
	return(0);
}



//*****************************************************************************
static int _ao_sw_trigger(dev_data_t* dev, void* arg)
{
	#undef	MASK
	#undef	SHIFT
	#define	MASK	(0x1L << SHIFT)
	#define	SHIFT	23

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bcr_32, MASK, MASK);
	return(0);
}



//*****************************************************************************
static int _ao_thresh_lvl(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x7FFFF, dev->vaddr.gsc_obtr_32, 18, 0);
	return(ret);
}



//*****************************************************************************
static int _ao_thresh_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_THRESH_STS_CLEAR,
		AISS16AO2_THRESH_STS_SET,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_obtr_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _data_format(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_DATA_FORMAT_2S_COMP,
		AISS16AO2_DATA_FORMAT_OFF_BIN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 25, 25);
	return(ret);
}



//*****************************************************************************
static int _dio_dir_out(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x3F, dev->vaddr.gsc_diopr_32, 13, 8);
	return(ret);
}



//*****************************************************************************
static int _dio_read(dev_data_t* dev, s32* arg)
{
	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_range_reg(dev, arg, 0, 0x3F, dev->vaddr.gsc_diopr_32, 5, 0);
	return(ret);
}



//*****************************************************************************
static int _dio_write(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x3F, dev->vaddr.gsc_diopr_32, 5, 0);
	return(ret);
}



//*****************************************************************************
static int _gen_a_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_GEN_ENABLE_NO,
		AISS16AO2_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 26, 26);
	return(ret);
}



//*****************************************************************************
static int _gen_a_ndiv(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.rate_gen_ndiv_min,
			dev->cache.rate_gen_ndiv_max,
			dev->vaddr.gsc_rgar_32,
			23,
			0);
	return(ret);
}



//*****************************************************************************
static int _gen_b_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_GEN_ENABLE_NO,
		AISS16AO2_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 27, 27);
	return(ret);
}



//*****************************************************************************
static int _gen_b_ndiv(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.rate_gen_ndiv_min,
			dev->cache.rate_gen_ndiv_max,
			dev->vaddr.gsc_rgbr_32,
			23,
			0);
	return(ret);
}



//*****************************************************************************
static int _gen_c_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_GEN_ENABLE_NO,
		AISS16AO2_GEN_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 22, 22);
	return(ret);
}



//*****************************************************************************
static int _gen_c_ndiv(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(
			dev,
			arg,
			dev->cache.rate_gen_ndiv_min,
			dev->cache.rate_gen_ndiv_max,
			dev->vaddr.gsc_rgcr_32,
			23,
			0);
	return(ret);
}



//*****************************************************************************
static int _irq_enable(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0x7FFF, dev->vaddr.gsc_psr_32, 14, 0);
	return(ret);
}



//*****************************************************************************
static int _master_clock(dev_data_t* dev, s32* arg)
{
	int	ret		= 0;

	if (arg[0] == -1)
	{
		arg[0]	= dev->cache.master_clock;
	}
	else if (	(arg[0] >= AISS16AO2_MASTER_CLOCK_MIN) &&
				(arg[0] <= AISS16AO2_MASTER_CLOCK_MAX))
	{
		dev->cache.master_clock			= arg[0];
		dev->cache.rate_gen_ndiv_min	= dev->cache.master_clock / dev->cache.fsamp_max;
	}
	else
	{
		ret	= -EINVAL;
	}

	return(ret);
}



//*****************************************************************************
static int _trigger_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		AISS16AO2_TRIGGER_MODE_IN,
		AISS16AO2_TRIGGER_MODE_OUT,
		-1	// terminate list
	};

	int	ret;

	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bcr_32, 21, 21);
	return(ret);
}



//*****************************************************************************
static int _rx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_read_abort_active_xfer(dev, &dev->io.rx);
	return(0);
}



//*****************************************************************************
static int _tx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_write_abort_active_xfer(dev, &dev->io.tx);
	return(0);
}



// variables ******************************************************************

const gsc_ioctl_t	dev_ioctl_list[]	=
{
	{ AISS16AO2_IOCTL_REG_READ,			(void*) gsc_reg_read_ioctl		},
	{ AISS16AO2_IOCTL_REG_WRITE,		(void*) gsc_reg_write_ioctl		},
	{ AISS16AO2_IOCTL_REG_MOD,			(void*) gsc_reg_mod_ioctl		},
	{ AISS16AO2_IOCTL_QUERY,			(void*) _query					},
	{ AISS16AO2_IOCTL_INITIALIZE,		(void*) initialize_ioctl		},
	{ AISS16AO2_IOCTL_AUTO_CALIBRATE,	(void*) _auto_calibrate			},
	{ AISS16AO2_IOCTL_AUTO_CAL_STS,		(void*) _auto_cal_sts			},
	{ AISS16AO2_IOCTL_RX_IO_MODE,		(void*) _rx_io_mode				},
	{ AISS16AO2_IOCTL_RX_IO_OVERFLOW,	(void*) _rx_io_overflow			},
	{ AISS16AO2_IOCTL_RX_IO_TIMEOUT,	(void*) _rx_io_timeout			},
	{ AISS16AO2_IOCTL_RX_IO_UNDERFLOW,	(void*) _rx_io_underflow		},
	{ AISS16AO2_IOCTL_TX_IO_MODE,		(void*) _tx_io_mode				},
	{ AISS16AO2_IOCTL_TX_IO_OVERFLOW,	(void*) _tx_io_overflow			},
	{ AISS16AO2_IOCTL_TX_IO_TIMEOUT,	(void*) _tx_io_timeout			},
	{ AISS16AO2_IOCTL_AI_BUF_CLEAR,		(void*) _ai_buf_clear			},
	{ AISS16AO2_IOCTL_AI_BUF_LEVEL,		(void*) _ai_buf_level			},
	{ AISS16AO2_IOCTL_AI_BUF_OVERFLOW,	(void*) _ai_buf_overflow		},
	{ AISS16AO2_IOCTL_AI_BUF_UNDERFLOW,	(void*) _ai_buf_underflow		},
	{ AISS16AO2_IOCTL_AI_BURST_ENABLE,	(void*) _ai_burst_enable		},
	{ AISS16AO2_IOCTL_AI_BURST_SIZE,	(void*) _ai_burst_size			},
	{ AISS16AO2_IOCTL_AI_BURST_STATUS,	(void*) _ai_burst_status		},
	{ AISS16AO2_IOCTL_AI_CHAN_SEL,		(void*) _ai_chan_sel			},
	{ AISS16AO2_IOCTL_AI_CLOCK_MODE,	(void*) _ai_clock_mode			},
	{ AISS16AO2_IOCTL_AI_ENABLE,		(void*) _ai_enable				},
	{ AISS16AO2_IOCTL_AI_FSAMP_RANGE,	(void*) _ai_fsamp_range			},
	{ AISS16AO2_IOCTL_AI_MODE,			(void*) _ai_mode				},
	{ AISS16AO2_IOCTL_AI_RANGE_A,		(void*) _ai_range_a				},
	{ AISS16AO2_IOCTL_AI_RANGE_B,		(void*) _ai_range_b				},
	{ AISS16AO2_IOCTL_AI_SW_CLOCK,		(void*) _ai_sw_clock			},
	{ AISS16AO2_IOCTL_AI_SW_TRIGGER,	(void*) _ai_sw_trigger			},
	{ AISS16AO2_IOCTL_AI_THRESH_LVL,	(void*) _ai_thresh_lvl			},
	{ AISS16AO2_IOCTL_AI_THRESH_STS,	(void*) _ai_thresh_sts			},
	{ AISS16AO2_IOCTL_AO_BUF_CLEAR,		(void*) _ao_buf_clear			},
	{ AISS16AO2_IOCTL_AO_BUF_LEVEL,		(void*) _ao_buf_level			},
	{ AISS16AO2_IOCTL_AO_BUF_LOAD_REQ,	(void*) _ao_load_req			},
	{ AISS16AO2_IOCTL_AO_BUF_LOAD_STS,	(void*) _ao_load_sts			},
	{ AISS16AO2_IOCTL_AO_BUF_MODE,		(void*) _ao_buf_mode			},
	{ AISS16AO2_IOCTL_AO_BUF_OVER_DATA,	(void*) _ao_buf_over_data		},
	{ AISS16AO2_IOCTL_AO_BUF_OVER_FRAM,	(void*) _ao_buf_over_fram		},
	{ AISS16AO2_IOCTL_AO_BURST_ENABLE,	(void*) _ao_burst_enable		},
	{ AISS16AO2_IOCTL_AO_BURST_STATUS,	(void*) _ao_burst_status		},
	{ AISS16AO2_IOCTL_AO_CHAN_SEL,		(void*) _ao_chan_sel			},
	{ AISS16AO2_IOCTL_AO_CLOCK_MODE,	(void*) _ao_clock_mode			},
	{ AISS16AO2_IOCTL_AO_CLOCK_STATUS,	(void*) _ao_clock_status		},
	{ AISS16AO2_IOCTL_AO_ENABLE,		(void*) _ao_enable				},
	{ AISS16AO2_IOCTL_AO_MODE,			(void*) _ao_mode				},
	{ AISS16AO2_IOCTL_AO_OUTPUT_MODE,	(void*) _ao_output_mode			},
	{ AISS16AO2_IOCTL_AO_RANGE,			(void*) _ao_range				},
	{ AISS16AO2_IOCTL_AO_SW_CLOCK,		(void*) _ao_sw_clock			},
	{ AISS16AO2_IOCTL_AO_SW_TRIGGER,	(void*) _ao_sw_trigger			},
	{ AISS16AO2_IOCTL_AO_THRESH_LVL,	(void*) _ao_thresh_lvl			},
	{ AISS16AO2_IOCTL_AO_THRESH_STS,	(void*) _ao_thresh_sts			},
	{ AISS16AO2_IOCTL_DATA_FORMAT,		(void*) _data_format			},
	{ AISS16AO2_IOCTL_DIO_DIR_OUT,		(void*) _dio_dir_out			},
	{ AISS16AO2_IOCTL_DIO_READ,			(void*) _dio_read				},
	{ AISS16AO2_IOCTL_DIO_WRITE,		(void*) _dio_write				},
	{ AISS16AO2_IOCTL_GEN_A_ENABLE,		(void*) _gen_a_enable			},
	{ AISS16AO2_IOCTL_GEN_A_NDIV,		(void*) _gen_a_ndiv				},
	{ AISS16AO2_IOCTL_GEN_B_ENABLE,		(void*) _gen_b_enable			},
	{ AISS16AO2_IOCTL_GEN_B_NDIV,		(void*) _gen_b_ndiv				},
	{ AISS16AO2_IOCTL_GEN_C_ENABLE,		(void*) _gen_c_enable			},
	{ AISS16AO2_IOCTL_GEN_C_NDIV,		(void*) _gen_c_ndiv				},
	{ AISS16AO2_IOCTL_IRQ_ENABLE,		(void*) _irq_enable				},
	{ AISS16AO2_IOCTL_MASTER_CLOCK,		(void*) _master_clock			},
	{ AISS16AO2_IOCTL_TRIGGER_MODE,		(void*) _trigger_mode			},
	{ AISS16AO2_IOCTL_RX_IO_ABORT,		(void*) _rx_io_abort			},
	{ AISS16AO2_IOCTL_TX_IO_ABORT,		(void*) _tx_io_abort			},
	{ AISS16AO2_IOCTL_WAIT_EVENT,		(void*) gsc_wait_event_ioctl	},
	{ AISS16AO2_IOCTL_WAIT_CANCEL,		(void*) gsc_wait_cancel_ioctl	},
	{ AISS16AO2_IOCTL_WAIT_STATUS,		(void*) gsc_wait_status_ioctl	},
	{ -1, NULL }
};


